==> huge-play-26.3.0-2.tgz: package/middleware/reference_in_pouchdb.coffee.md <==
Register the database we use to store session state.
FIXME: use redis instead.

    @name = 'huge-play:middleware:reference_in_pouchdb'
    debug = require('tangible') @name

    seem = require 'seem'
    PouchDB = require 'shimaore-pouchdb'
    uuidV4 = require 'uuid/v4'

    sleep = (timeout) ->
      new Promise (resolve) ->
        setTimeout resolve, timeout

    @server_pre = ->

* cfg.session.base (URI) base URI for databases that store automated / complete call records (call-center oriented). Default: cfg.data.url

      base = @cfg.session?.base ? @cfg.data?.url

      unless base
        debug.dev 'No cfg.session.base nor cfg.data.url, references will not be saved.'
        return

      base += '/' unless base.match /\/$/

* cfg.REFERENCE_DB_PREFIX (string) database-name prefix for references. Current value: `reference`.
(Name is fixed because it also appears in `spicy-action/public_proxy`.)

      db_prefix = @cfg.REFERENCE_DB_PREFIX = 'reference'

      if @cfg.get_reference_data? or @cfg.update_reference_data?
        debug.dev 'Another module provided the functions, not starting.'
        return

      RemotePouchDB = PouchDB.defaults prefix: base

      current_db_name = null
      current_db = null

      get_db = (database) ->
        if current_db_name is database
          current_db
        else
          current_db?.close()
          current_db_name = database
          current_db = new RemotePouchDB current_db_name

      name_for_id = (id) ->
        period = id.substring 0, 7
        database = [db_prefix,period].join '-'

      now = ->
        new Date().toJSON()

      {host} = @cfg

Get
---

      @cfg.get_reference_data = get_data = (id) ->
        database = name_for_id id

        db = get_db database
        db
          .get id
          .catch -> _id:id

Update
------

### Report

Normally reports are generated in-call and stored in the `reports` array of the call.
For out-of-call reports we store them in a structure similar to the one used for calls.

      @cfg.save_reports = save_reports = (reports,tries = 3) ->

        timestamp = now()
        reference = null

        reports.forEach (report) ->
          report.timestamp ?= timestamp
          report.host = host
          report.type = 'report'
          reference ?= report.reference if report.reference?

        reference ?= timestamp

        database = name_for_id reference
        db = get_db database

        db
        .bulkDocs reports

In case of failure, retry.
FIXME Properly handle bulkDocs semantics.

        .catch seem (error) =>
          debug "reference error: #{error.stack}", error
          if tries-- > 0
            yield sleep 181
            yield save_reports notification, tries
          else
            call

### Call/Session

A call/session is a single call handled by a FreeSwitch call to `socket`. It is slightly related but not exactly quite what SIP would call a "call".

      @cfg.update_call_data = save_call = seem (call_data,tries = 3) ->

        # assert call_data.reference?
        # assert call_data.session?

        database = name_for_id call_data.reference

        db = get_db database
        doc = yield db.get(call_data.session).catch -> null

        doc ?=
          _id: call_data.session
          timestamp: now()
          host: host
          type: 'call'

        for own k,v of call_data when k[0] isnt '_' and v? and typeof v isnt 'function'
          doc[k] = v

        db
        .put doc

In case of success, return the updated document.

        .then ({rev}) ->
          doc._rev = rev
          doc

In case of failure, retry, or return the submitted data.

        .catch seem (error) =>
          debug "update_call_data error: #{error.stack}", error
          if tries-- > 0
            yield sleep 179
            yield save_call call_data, tries
          else
            call_data

### Reference

A reference is a what a human would call a `call`: a call originated somewhere, as it progresses through menus, redirections, transfers, …
The main purpose of storing the reference-data is to allow data to be propagated along the chain.

      @cfg.update_reference_data = save_data = seem (reference_data,tries = 3) ->
        {_id} = reference_data
        database = name_for_id _id

        db = get_db database
        doc = yield db.get(_id).catch -> null

        doc ?=
          _id: _id
          timestamp: now()
          host: host
          type: 'reference'

Merge tags (but keep them ordered)

        doc.tags ?= []
        if reference_data.tags?
          tags = new Set doc.tags
          for tag in reference_data.tags when not tags.has tag
            doc.tags.push tag

Known fields are:
- tags (managed above)
- destination (used by exultant-songs)
- account (set by huge-play)

        for own k,v of reference_data when k[0] isnt '_' and k isnt 'tags' and v? and typeof v isnt 'function'
          doc[k] = v

        db
        .put doc

In case of success, return the updated document.

        .then ({rev}) ->
          doc._rev = rev
          doc

In case of failure, retry, or return the submitted data.

        .catch seem (error) =>
          debug "update_reference_data error: #{error.stack}", error
          if tries-- > 0
            yield sleep 173
            yield save_data reference_data, tries
          else
            reference_data

      debug 'Ready.'

    @include = ->

==> huge-play-27.0.0-4.tgz: package/middleware/reference_in_pouchdb.coffee.md <==
Register the database we use to store session state.
FIXME: use redis instead.

    @name = 'huge-play:middleware:reference_in_pouchdb'
    debug = require('tangible') @name

    seem = require 'seem'
    PouchDB = require 'shimaore-pouchdb'
    uuidV4 = require 'uuid/v4'

    sleep = (timeout) ->
      new Promise (resolve) ->
        setTimeout resolve, timeout

    update_doc = (doc,data) ->

Copy the `_in` field (which CouchDB won't accept) in a properly named one.

      if data._in?
        doc.in ?= data._in

Merge tags (but keep them ordered)

      if data.tags?
        doc.tags ?= []
        doc_tags = new Set doc.tags
        for tag in data.tags when not doc_tags.has tag
          doc.tags.push tag

      for own k,v of data when k[0] isnt '_' and k isnt 'tags' and v? and typeof v isnt 'function'
        doc[k] = v

      doc

    @notify = ->

* cfg.session.base (URI) base URI for databases that store automated / complete call records (call-center oriented). Default: cfg.data.url

      base = @cfg.session?.base ? @cfg.data?.url

      unless base
        debug.dev 'No cfg.session.base nor cfg.data.url, references will not be saved.'
        return

      base += '/' unless base.match /\/$/

* cfg.REFERENCE_DB_PREFIX (string) database-name prefix for references. Current value: `reference`.
(Name is fixed because it also appears in `spicy-action/public_proxy`.)

      db_prefix = @cfg.REFERENCE_DB_PREFIX = 'reference'

      RemotePouchDB = PouchDB.defaults prefix: base

      current_db_name = null
      current_db = null

      get_db = (database) ->
        if current_db_name is database
          current_db
        else
          current_db?.close()
          current_db_name = database
          current_db = new RemotePouchDB current_db_name

      name_for_id = (id) ->
        period = id.substring 0, 7
        database = [db_prefix,period].join '-'

      now = ->
        new Date().toJSON()

      {host} = @cfg

Get
---

      @cfg.get_reference_data = get_data = (id) ->
        database = name_for_id id

        db = get_db database
        db
          .get id
          .catch -> _id:id

Update
------

### Report

Normally reports are generated in-call and stored in the `reports` array of the call.
For out-of-call reports we store them in a structure similar to the one used for calls.

      save_reports = (reports,tries = 3) ->

        timestamp = now()
        reference = reports.find( (r) -> r.reference? )?.reference
        reference ?= timestamp

        database = name_for_id reference
        db = get_db database

        docs = reports.map (report) -> update_doc {}, report

        db
        .bulkDocs docs

In case of failure, retry.
FIXME Properly handle bulkDocs semantics.

        .catch seem (error) =>
          debug "reference error: #{error.stack}", error
          if tries-- > 0
            yield sleep 181
            yield save_reports reports, tries
          else
            call

      @cfg.statistics.on 'reports', save_reports
      @cfg.statistics.on 'report', (report) ->
        save_reports [report]

### Call/Session

A call/session is a single call handled by a FreeSwitch call to `socket`. It is slightly related but not exactly quite what SIP would call a "call".

      save_call = seem (call_data,tries = 3) ->

        # assert call_data.reference?
        # assert call_data.session?

        database = name_for_id call_data.reference

        db = get_db database
        doc = yield db.get(call_data.session).catch -> null

        doc ?=
          _id: call_data.session

        doc = update_doc doc, call_data

        db
        .put doc

In case of success, return the updated document.

        .then ({rev}) ->
          doc._rev = rev
          doc

In case of failure, retry, or return the submitted data.

        .catch seem (error) =>
          debug "update_call_data error: #{error.stack}", error
          if tries-- > 0
            yield sleep 179
            yield save_call call_data, tries
          else
            call_data

      @cfg.statistics.on 'call', save_call

### Reference

A reference is a what a human would call a `call`: a call originated somewhere, as it progresses through menus, redirections, transfers, …
The main purpose of storing the reference-data is to allow data to be propagated along the chain.

      save_data = seem (reference_data,tries = 3) ->
        {_id} = reference_data
        database = name_for_id _id

        db = get_db database
        doc = yield db.get(_id).catch -> null

        doc ?=
          _id: _id

        doc = update_doc doc, reference_data

        db
        .put doc

In case of success, return the updated document.

        .then ({rev}) ->
          doc._rev = rev
          doc

In case of failure, retry, or return the submitted data.

        .catch seem (error) =>
          debug "update_reference_data error: #{error.stack}", error
          if tries-- > 0
            yield sleep 173
            yield save_data reference_data, tries
          else
            reference_data

      debug 'Ready.'

      @cfg.statistics.on 'reference', save_data

    @include = ->
